﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using CG2.Lighting;
using CG2.Texturing;

namespace CG2.Shading
{
	public class PhongShader : Shader
	{
		#region Properties

		public Sampler DiffuseSampler = new ColorSampler(new Vector4(1.0, 1.0, 1.0));
		public Sampler SpecularSampler = new ColorSampler(new Vector4(1.0, 1.0, 1.0));
		public Sampler AmbientSampler = new ColorSampler(new Vector4(0.1, 0.1, 0.1));

		public Sampler NormalSampler = null;
		public Boolean IsDeltaNormalSampler = false;

		public Double Shininess = 20.0;

		#endregion


		#region Init

		public PhongShader()
		{
		}

		public PhongShader(Vector4 diffuseColor)
		{
			DiffuseSampler = new ColorSampler(diffuseColor);
		}

		public PhongShader(Vector4 diffuseColor, Vector4 specularColor)
		{
			DiffuseSampler = new ColorSampler(diffuseColor);
			SpecularSampler = new ColorSampler(specularColor);
		}

		public PhongShader(Vector4 diffuseColor, Vector4 specularColor, Vector4 ambientColor)
		{
			DiffuseSampler = new ColorSampler(diffuseColor);
			SpecularSampler = new ColorSampler(specularColor);
			AmbientSampler = new ColorSampler(ambientColor);
		}

		#endregion


		#region Shading

		public override Vector4 GetNormal(Vector4 point, Vector4 normal)
		{
			if (NormalSampler == null) return normal;

			Vector4 n = NormalSampler.GetSample(point);
			if (IsDeltaNormalSampler) n += normal;
			Double nl = n.Length;

			return (nl == 0) ? normal : 1 / nl * n;
		}

		/// <summary>
		/// Blinn-Phong local shading model
		/// </summary>
		/// <param name="point"></param>
		/// <param name="normal"></param>
		/// <param name="eyeDir"></param>
		/// <param name="lightDir"></param>
		/// <param name="lightColor"></param>
		/// <returns></returns>
		public override Vector4 GetColor(Vector4 point, Vector4 normal, Vector4 eyeDir, Vector4 lightDir, Double lightIntensity, Light light)
		{
			Double diffuseFactor = lightIntensity * (normal * lightDir);
			if (diffuseFactor < 0) diffuseFactor = 0;

			Vector4 half = (lightDir + eyeDir).Normalized;
			Double specularFactor = lightIntensity * Math.Pow(normal * half, Shininess);

			Vector4 diffuseColor = DiffuseSampler.GetSample(point);
			Vector4 specularColor = SpecularSampler.GetSample(point);
			Vector4 ambientColor = AmbientSampler.GetSample(point);

			Vector4 color = new Vector4();

			color = diffuseFactor * (diffuseColor ^ light.DiffuseColor);
			color += specularFactor * (specularColor ^ light.SpecularColor);
			color += ambientColor ^ light.AmbientColor;

			return color;
		}

		#endregion
	}
}
